// Conway's Game of Life (see conwaylife.com)
//
// Runs on Waveshare RP2040-PiZero
// output on DVI (HDMI) port
// (see HackerBoxes.com -> HackerBox 0112)
//

#include <PicoDVI.h>

DVIGFX1 display(DVI_RES_640x480p60, true, waveshare_rp2040_pizero);

const int w = 320;
const int h = 240;
unsigned char next[w][h];
unsigned char curr[w][h];


void setup() {
  randomSeed(analogRead(A0));
  display.begin();
  init_blank();          // clear curr matrix
  //init_random();       // initialize curr matrix with noise
  init_multi_glider(50); // initialize curr matrix with 50 gliders
}

void init_random(){
  int x, y;
  for (x=0; x<320; x++) {
    for (y=0; y<240; y++) {
      if (random(100) > 70) {  // make 30% of cells alive
        curr[x][y] = 1;
      }
      else {
        curr[x][y] = 0;
      }
    }
  }
}

void init_blank() {
  int i, j;
  for (i = 0; i < 320; i++) {
    for (j = 0; j < 240; j++) {
      curr[i][j] = 0;
    }
  }
}

void init_glider(int x, int y){
  curr[x+1][y] = 1;
  curr[x+2][y+1] = 1;
  curr[x][y+2] = 1;
  curr[x+1][y+2] = 1;
  curr[x+2][y+2] = 1;
}

void init_multi_glider(int n) {
  for (int i=0; i < n; i++)   
    init_glider(random(w), random(h));
}

void display_curr() {
  int x, y;
  display.fillScreen(0); // clear back framebuffer
  for (x=0; x<320; x++) {
    for (y=0; y<240; y++) {
      if (curr[x][y] == 1) {
        display.drawPixel(x*2, y*2, 0xFFFF);
        display.drawPixel(x*2+1, y*2, 0xFFFF);
        display.drawPixel(x*2, y*2+1, 0xFFFF);
        display.drawPixel(x*2+1, y*2+1, 0xFFFF);
      }
    }
  }
  display.swap();  // swap front/back framebuffers
}

void invert() {
  int x, y;
  for (x=0; x<320; x++) {
    for (y=0; y<240; y++) {
      if (curr[x][y] == 1) {
        next[x][y] = 0;
      }
      else {
        next[x][y] = 1;
      }
    }
  }
  for (int i=0; i < w; i++){
    for (int j=0; j < h; j++) {
      curr[i][j] = next[i][j];
    }
  }
}

int count_neighbors(int x, int y){
  int count = 0;
  int dx;
  int dy;
  int test_x;
  int test_y;
  for (dx = -1; dx <= 1; dx++) {
    for (dy = -1; dy <= 1; dy++) {
      test_x = x + dx;
      test_y = y + dy;
      if (test_x < 0)
        test_x = w-1;
      if (test_x >= w)
        test_x = 0;
      if (test_y < 0)
        test_y = h-1;
      if (test_y >= h)
        test_y = 0;
      if (curr[test_x][test_y])
        count += 1;
    }
  }
  if (curr[x][y])
    count -= 1;
  return count;
}

void evolve(){
  int n, i, j;
  for (i=0; i < w; i++){
    for (j=0; j < h; j++) {
      n = count_neighbors(i, j);
      // Any live cell with fewer than two live neighbors dies
      if ((curr[i][j] == 1) && (n < 2))
        next[i][j] = 0;
      // Any live cell with two or three live neighbors remains alive
      if ((curr[i][j] == 1) && (n==3))
        next[i][j] = 1;
      if ((curr[i][j] == 1) && (n==2))
        next[i][j] = 1;
      // Any live cell with more than three live neighbors dies
      if ((curr[i][j] == 1) && (n > 3))
        next[i][j] = 0;
      // Any dead cell with exactly three live neighbors becomes alive
      if ((curr[i][j] == 0) && (n == 3))
        next[i][j] = 1;
      if ((curr[i][j] == 0) && (n != 3))
        next[i][j] = 0;
    }
  }        
  for (i=0; i < w; i++){
    for (j=0; j < h; j++) {
      curr[i][j] = next[i][j];
    }
  }
}

void loop() {
  display_curr();
  evolve();
}
